/* obfc.c
 *
 * main program of the optimizing brainfuck compiler
 *
 * Copyright (C) 2004 Manuel Schiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "bf.h"
#include "bfparse.h"
#include "loopopt.h"
#include "be.h"

static char *verstr = "obfc version 0.0.2";

/* list of all supported back ends */
static struct be_list_t {
	char *name; /* a back end name */
	int (*generate)(FILE *out, bfinstr_t *pgm); /* code generator */
} be_list[5] = {
	{ "c-int", gen_C_int },
	{ "c-char", gen_C_char },
	{ "x86-32", gen_x86_32 },
	{ "x86-8", gen_x86_8 },
	{ NULL, NULL }
};

/* return the index into the list of backends */
static int get_be(char *bename)
{
	int i;
	
	for (i = 0; be_list[i].name != NULL; i++) {
		if (strcasecmp(bename, be_list[i].name) == 0)
			return i;
	}

	fprintf(stderr, "obfc: unknown back end: %s\n", bename);
	exit(-1);
}

/* obvious... */
static void version()
{
	printf("%s\n", verstr);
	exit(0);
}

/* obvious as well... */
static void help()
{
	int i;
	
	printf("usage: obfc [-h|-v] [-b backend] -o outfile infile\n\n" \
		"\t-h\tgives help\n" \
		"\t-v\tshow version number\n" \
		"\t-o name\tspecify output file name\n" \
		"\t-b be\tspecify backend to use\n\n" \
		"available backends are:");
	for (i = 0; be_list[i].name != NULL; i++)
		printf(" %s", be_list[i].name);
	putchar('\n');
	exit(0);
}

int main(int argc, char **argv)
{
	long tmp;
	int c = 0, be_id = 0;
	char *iname = NULL, *oname = NULL;
	bfinstr_t *pgm = NULL;
	FILE *f;

	while ((c = getopt(argc, argv, "o:b:hv")) != -1) {
		switch (c) {
		case 'o':
			oname = optarg;
			break;
		case 'b':
			be_id = get_be(optarg);
			break;
		case 'h':
			help();
		case 'v':
			version();
		case '?':
			fprintf(stderr, "obfc: unknown option: -%c\n", optopt);
		default:
			break;
		}	
	}

	/* make sure we have exactly one input file */
	if (optind >= argc) {
		fprintf(stderr, "obfc: missing input file name.\n");
		exit(-2);
	}
	if (optind != (argc - 1)) {
		fprintf(stderr, "obfc: extra input file name.\n");
		exit(-3);
	}
	iname = argv[optind];
	if (oname == NULL) {
		fprintf(stderr, "obfc: no output file name specified.\n");
		exit(-4);
	}
	
	/* open input file */
	if ((f = fopen(iname, "r")) == NULL) {
		perror("obfc: can't open input file:");
		exit(-5);
	}

	pgm = parser(f); /* parse the input*/
	fclose(f); /* close input file */
	tmp = loopopt(pgm, 0); /* loop optimisation pass */
	
	/* open output file */
	if ((f = fopen(oname, "w")) == NULL) {
		perror("obfc: can't open output file:");
		exit(-6);
	}

	/* generate the code */	
	be_list[be_id].generate(f, pgm);

	free(pgm);
	fclose(f);
	return 0;
}
